from typing import Literal

TrackQuality = Literal["LOW", "HIGH", "LOSSLESS", "HI_RES_LOSSLESS"]
TrackArg = Literal["low", "normal", "high", "master"]
SinglesFilter = Literal["none", "only", "include"]

ARG_TO_QUALITY: dict[TrackArg, TrackQuality] = {
    "low": "LOW",
    "normal": "HIGH",
    "high": "LOSSLESS",
    "master": "HI_RES_LOSSLESS",
}

QUALITY_TO_ARG = {v: k for k, v in ARG_TO_QUALITY.items()}



class AuthError(Exception):
    def __init__(
        self, status: int, error: str, sub_status: str, error_description: str
    ):
        self.status = status
        self.error = error
        self.sub_status = sub_status
        self.error_description = error_description

    def __str__(self):
        return f"{self.status}: {self.error} - {self.error_description}"


class ApiError(Exception):
    def __init__(self, status: int, subStatus: str, userMessage: str):
        self.status = status
        self.sub_status = subStatus
        self.user_message = userMessage

    def __str__(self):
        return f"{self.user_message} ({self.status} - {self.sub_status})"



import click
import json

from io import TextIOWrapper
from os.path import splitext

from tiddl.utils import TidalResource
from tiddl.cli.ctx import Context, passContext


@click.group("file")
@click.argument("filename", type=click.File(mode="r"))
@passContext
def FileGroup(ctx: Context, filename: TextIOWrapper):
    """Parse txt or JSON file with urls."""

    _, extension = splitext(filename.name)

    resource_strings: list[str]

    match extension:
        case ".json":
            try:
                resource_strings = json.load(filename)
            except json.JSONDecodeError as e:
                raise click.UsageError(f"Cant decode JSON file - {e.msg}")

        case ".txt":
            resource_strings = [line.strip() for line in filename.readlines()]

        case _:
            raise click.UsageError(f"Unsupported file extension - {extension}")

    for string in resource_strings:
        try:
            ctx.obj.resources.append(TidalResource.fromString(string))
        except ValueError as e:
            click.echo(click.style(e, "red"))

    click.echo(click.style(f"Loaded {len(ctx.obj.resources)} resources", "green"))



import unittest

from tiddl.models.resource import Track
from tiddl.utils import TidalResource, formatTrack


class TestTidalResource(unittest.TestCase):
    def test_resource_parsing(self):
        positive_cases = [
            ("https://tidal.com/browse/track/12345678", "track", "12345678"),
            ("track/12345678", "track", "12345678"),
            ("https://tidal.com/browse/video/12345678", "video", "12345678"),
            ("video/12345678", "video", "12345678"),
            ("https://tidal.com/browse/album/12345678", "album", "12345678"),
            ("album/12345678", "album", "12345678"),
            ("https://tidal.com/browse/playlist/12345678", "playlist", "12345678"),
            ("playlist/12345678", "playlist", "12345678"),
            ("https://tidal.com/browse/artist/12345678", "artist", "12345678"),
            ("artist/12345678", "artist", "12345678"),
            (
                "https://tidal.com/browse/mix/f93b015796bf93b015796b",
                "mix",
                "f93b015796bf93b015796b",
            ),
            ("mix/f93b015796bf93b015796b", "mix", "f93b015796bf93b015796b"),
        ]

        for resource, expected_type, expected_id in positive_cases:
            with self.subTest(resource=resource):
                tidal_resource = TidalResource.fromString(resource)
                self.assertEqual(tidal_resource.type, expected_type)
                self.assertEqual(tidal_resource.id, expected_id)

    def test_failing_cases(self):
        failing_cases = [
            "https://tidal.com/browse/invalid/12345678",
            "invalid/12345678",
            "https://tidal.com/browse/track/invalid",
            "track/invalid",
            "",
            "invalid",
            "https://tidal.com/browse/track/",
            "track/",
            "/12345678",
        ]

        for resource in failing_cases:
            with self.subTest(resource=resource):
                with self.assertRaises(ValueError):
                    TidalResource.fromString(resource)


class TestFormatTrack(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.track = Track(
            **{
                "id": 66421438,
                "title": "Shutdown",
                "duration": 189,
                "replayGain": -9.95,
                "peak": 0.966051,
                "allowStreaming": True,
                "streamReady": True,
                "adSupportedStreamReady": True,
                "djReady": True,
                "stemReady": False,
                "streamStartDate": "2016-11-15T00:00:00.000+0000",
                "premiumStreamingOnly": False,
                "trackNumber": 9,
                "volumeNumber": 1,
                "version": None,
                "popularity": 24,
                "copyright": "(P) 2016 Boy Better Know",
                "bpm": 69,
                "url": "http://www.tidal.com/track/66421438",
                "isrc": "GB7QY1500024",
                "editable": False,
                "explicit": True,
                "audioQuality": "LOSSLESS",
                "audioModes": ["STEREO"],
                "mediaMetadata": {"tags": ["LOSSLESS", "HIRES_LOSSLESS"]},
                "artist": {
                    "id": 3566984,
                    "name": "Skepta",
                    "type": "MAIN",
                    "picture": "747af850-fa9c-4178-a3e6-49259b67df86",
                },
                "artists": [
                    {
                        "id": 3566984,
                        "name": "Skepta",
                        "type": "MAIN",
                        "picture": "747af850-fa9c-4178-a3e6-49259b67df86",
                    }
                ],
                "album": {
                    "id": 66421429,
                    "title": "Konnichiwa",
                    "cover": "e0c2f05e-e21f-47c5-9c37-2993437df27d",
                    "vibrantColor": "#ae3b31",
                    "videoCover": None,
                },
                "mixes": {"TRACK_MIX": "001aa4abeb471e8f55f5784772b478"},
                "playlistNumber": None,
            }
        )

    def test_templating(self):
        test_cases = [
            ("{id}", "66421438"),
            ("{title}", "Shutdown"),
            ("{version}", ""),
            ("{artist}", "Skepta"),
            ("{artists}", "Skepta"),
            ("{album}", "Konnichiwa"),
            ("{number}", "9"),
            ("{disc}", "1"),
            ("{date:%m-%d-%y}", "11-15-16"),
            ("{date:%Y}", "2016"),
            ("{year}", "2016"),
            ("{playlist_number}", "0"),
            ("{playlist_number:02d}", "00"),
            ("{bpm}", "69"),
            ("{quality}", "high"),
            ("{artist}/{album}/{title}", "Skepta/Konnichiwa/Shutdown"),
            ("{number:02d}. {title}", "09. Shutdown"),
        ]

        for template, expected_result in test_cases:
            with self.subTest(template=template, expected_result=expected_result):
                result = formatTrack(template, self.track)
                self.assertEqual(result, expected_result)

    def test_invalid_characters(self):
        test_cases = ["\\", ":", '"', "?", "<", ">", "|", "{number}:{title}", "{date}"]

        for template in test_cases:
            with self.subTest(template=template):
                with self.assertRaises(ValueError):
                    formatTrack(template, self.track)


if __name__ == "__main__":
    unittest.main()



import click

from tiddl.config import CONFIG_PATH
from tiddl.cli.ctx import Context, passContext


@click.command("config")
@click.option(
    "--open",
    "-o",
    "OPEN_CONFIG",
    is_flag=True,
    help="Open the configuration file with the default editor.",
)
@click.option(
    "--locate",
    "-l",
    "LOCATE_CONFIG",
    is_flag=True,
    help="Launch a file manager with the located configuration file.",
)
@click.option(
    "--print",
    "-p",
    "PRINT_CONFIG",
    is_flag=True,
    help="Show current configuration.",
)
@passContext
def ConfigCommand(
    ctx: Context, OPEN_CONFIG: bool, LOCATE_CONFIG: bool, PRINT_CONFIG: bool
):
    """
    Configuration file options.

    By default it prints location of tiddl config file.

    This command can be used in variable like `vim $(tiddl config)`
    - this will open your config with vim editor.
    """

    if OPEN_CONFIG:
        click.launch(str(CONFIG_PATH))

    elif LOCATE_CONFIG:
        click.launch(str(CONFIG_PATH), locate=True)

    elif PRINT_CONFIG:
        config_without_auth = ctx.obj.config.model_copy()
        del config_without_auth.auth
        ctx.obj.console.print(config_without_auth.model_dump_json(indent=2))

    else:
        click.echo(str(CONFIG_PATH))



import click

from tiddl.utils import TidalResource
from tiddl.cli.ctx import Context, passContext


class TidalURL(click.ParamType):
    def convert(self, value: str, param, ctx) -> TidalResource:
        try:
            return TidalResource.fromString(value)
        except ValueError as e:
            self.fail(message=str(e), param=param, ctx=ctx)


@click.group("url")
@click.argument("url", type=TidalURL())
@passContext
def UrlGroup(ctx: Context, url: TidalResource):
    """
    Get Tidal URL.

    It can be Tidal link or `resource_type/resource_id` format.
    The resource can be a track, video, album, playlist or artist.
    """

    ctx.obj.resources.append(url)



import unittest

from tiddl.config import Config
from tiddl.api import TidalApi


class TestApi(unittest.TestCase):
    api: TidalApi

    def setUp(self):
        config = Config.fromFile()
        auth = config.auth

        token, user_id, country_code = (
            auth.token,
            auth.user_id,
            auth.country_code,
        )

        assert token, "No token found in config file"
        assert user_id, "No user_id found in config file"
        assert country_code, "No country_code found in config file"

        self.api = TidalApi(token, user_id, country_code)

    def test_ready(self):
        session = self.api.getSession()

        self.assertEqual(session.userId, int(self.api.user_id))
        self.assertEqual(session.countryCode, self.api.country_code)

    def test_track(self):
        track = self.api.getTrack(103805726)
        self.assertEqual(track.title, "Stronger")

    def test_artist(self):
        artist = self.api.getArtist(25022)
        self.assertEqual(artist.name, "Kanye West")

    def test_artist_albums(self):
        self.api.getArtistAlbums(25022, filter="ALBUMS")
        self.api.getArtistAlbums(25022, filter="EPSANDSINGLES")

    def test_album(self):
        album = self.api.getAlbum(103805723)
        self.assertEqual(album.title, "Graduation")

    def test_album_items(self):
        album_items = self.api.getAlbumItems(103805723, limit=10)
        self.assertEqual(len(album_items.items), 10)

        album_items = self.api.getAlbumItems(103805723, limit=10, offset=10)
        self.assertEqual(len(album_items.items), 4)

    def test_album_items_credits(self):
        album_items = self.api.getAlbumItemsCredits(103805723, limit=10)
        self.assertEqual(len(album_items.items), 10)

    def test_playlist(self):
        playlist = self.api.getPlaylist("84974059-76af-406a-aede-ece2b78fa372")
        self.assertEqual(playlist.title, "Kanye West Essentials")

    def test_playlist_items(self):
        playlist_items = self.api.getPlaylistItems(
            "84974059-76af-406a-aede-ece2b78fa372"
        )
        self.assertEqual(len(playlist_items.items), 25)

    def test_favorites(self):
        favorites = self.api.getFavorites()
        self.assertGreaterEqual(len(favorites.PLAYLIST), 0)
        self.assertGreaterEqual(len(favorites.ALBUM), 0)
        self.assertGreaterEqual(len(favorites.VIDEO), 0)
        self.assertGreaterEqual(len(favorites.TRACK), 0)
        self.assertGreaterEqual(len(favorites.ARTIST), 0)

    def test_search(self):
        self.api.getSearch("Kanye West")

    def test_video(self):
        self.api.getVideo(373513584)

    def test_video_stream(self):
        self.api.getVideoStream(373513584)

    def test_lyrics(self):
        track_id = 103805726
        lyrics = self.api.getLyrics(track_id)
        self.assertEqual(lyrics.trackId, track_id)


if __name__ == "__main__":
    unittest.main()



from typing import Optional
from pydantic import BaseModel


class AuthUser(BaseModel):
    userId: int
    email: str
    countryCode: str
    fullName: Optional[str]
    firstName: Optional[str]
    lastName: Optional[str]
    nickname: Optional[str]
    username: str
    address: Optional[str]
    city: Optional[str]
    postalcode: Optional[str]
    usState: Optional[str]
    phoneNumber: Optional[str]
    birthday: Optional[int]
    channelId: int
    parentId: int
    acceptedEULA: bool
    created: int
    updated: int
    facebookUid: int
    appleUid: Optional[str]
    googleUid: Optional[str]
    accountLinkCreated: bool
    emailVerified: bool
    newUser: bool


class AuthResponse(BaseModel):
    user: AuthUser
    scope: str
    clientName: str
    token_type: str
    access_token: str
    expires_in: int
    user_id: int


class AuthResponseWithRefresh(AuthResponse):
    refresh_token: str


class AuthDeviceResponse(BaseModel):
    deviceCode: str
    userCode: str
    verificationUri: str
    verificationUriComplete: str
    expiresIn: int
    interval: int



import logging
import base64
from os import environ

from requests import request

from tiddl.exceptions import AuthError
from tiddl.models import auth

AUTH_URL = "https://auth.tidal.com/v1/oauth2"


def get_auth_credentials() -> tuple[str, str]:
    ENV_KEY = "TIDDL_AUTH"

    client_id, client_secret = (
        base64.b64decode(
            "ZlgySnhkbW50WldLMGl4VDsxTm45QWZEQWp4cmdKRkpiS05XTGVBeUtHVkdtSU51WFBQTEhWWEF2eEFnPQ=="
        )
        .decode()
        .split(";")
    )

    env_value = environ.get(ENV_KEY, None)

    if env_value:
        client_id, client_secret = env_value.split(";")

    return client_id, client_secret


CLIENT_ID, CLIENT_SECRET = get_auth_credentials()

logger = logging.getLogger(__name__)


def getDeviceAuth():
    req = request(
        "POST",
        f"{AUTH_URL}/device_authorization",
        data={"client_id": CLIENT_ID, "scope": "r_usr+w_usr+w_sub"},
    )

    data = req.json()

    if req.status_code == 200:
        return auth.AuthDeviceResponse(**data)

    raise AuthError(**data)


def getToken(device_code: str):
    req = request(
        "POST",
        f"{AUTH_URL}/token",
        data={
            "client_id": CLIENT_ID,
            "device_code": device_code,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "scope": "r_usr+w_usr+w_sub",
        },
        auth=(CLIENT_ID, CLIENT_SECRET),
    )

    data = req.json()

    if req.status_code == 200:
        return auth.AuthResponseWithRefresh(**data)

    raise AuthError(**data)


def refreshToken(refresh_token: str):
    req = request(
        "POST",
        f"{AUTH_URL}/token",
        data={
            "client_id": CLIENT_ID,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
            "scope": "r_usr+w_usr+w_sub",
        },
        auth=(CLIENT_ID, CLIENT_SECRET),
    )

    data = req.json()

    if req.status_code == 200:
        return auth.AuthResponse(**data)

    raise AuthError(**data)


def removeToken(access_token: str):
    req = request(
        "POST",
        "https://api.tidal.com/v1/logout",
        headers={"authorization": f"Bearer {access_token}"},
    )

    logger.debug((req.status_code, req.text))



import json
import logging
from pathlib import Path
from typing import Any, Literal, Type, TypeVar

from pydantic import BaseModel
from requests_cache import (
    CachedSession,
    EXPIRE_IMMEDIATELY,
    NEVER_EXPIRE,
    DO_NOT_CACHE,
)

from tiddl.models.api import (
    Album,
    AlbumItems,
    AlbumItemsCredits,
    Artist,
    ArtistAlbumsItems,
    Favorites,
    Playlist,
    PlaylistItems,
    Search,
    SessionResponse,
    Track,
    TrackStream,
    Video,
    VideoStream,
    Lyrics,
    MixItems,
)

from tiddl.models.constants import TrackQuality
from tiddl.exceptions import ApiError
from tiddl.config import HOME_PATH

DEBUG = False

T = TypeVar("T", bound=BaseModel)

logger = logging.getLogger(__name__)


def ensureLimit(limit: int, max_limit: int) -> int:
    if limit > max_limit:
        logger.warning(f"Max limit is {max_limit}")
        return max_limit

    return limit


class Limits:
    ARTIST_ALBUMS = 50
    ALBUM_ITEMS = 10
    ALBUM_ITEMS_MAX = 100
    PLAYLIST = 50
    MIX_ITEMS = 100


class TidalApi:
    URL = "https://api.tidal.com/v1"
    LIMITS = Limits

    def __init__(
        self, token: str, user_id: str, country_code: str, omit_cache=False
    ) -> None:
        self.user_id = user_id
        self.country_code = country_code

        # 3.0 TODO: change cache path
        CACHE_NAME = "tiddl_api_cache"

        self.session = CachedSession(
            cache_name=HOME_PATH / CACHE_NAME, always_revalidate=omit_cache
        )
        self.session.headers = {
            "Authorization": f"Bearer {token}",
            "Accept": "application/json",
        }

    def fetch(
        self,
        model: Type[T],
        endpoint: str,
        params: dict[str, Any] = {},
        expire_after=NEVER_EXPIRE,
    ) -> T:
        """Fetch data from the API and parse it into the given Pydantic model."""

        req = self.session.get(
            f"{self.URL}/{endpoint}", params=params, expire_after=expire_after
        )

        logger.debug(
            (
                endpoint,
                params,
                req.status_code,
                "HIT" if req.from_cache else "MISS",
            )
        )

        data = req.json()

        if DEBUG:
            debug_data = {
                "status_code": req.status_code,
                "endpoint": endpoint,
                "params": params,
                "data": data,
            }

            path = Path(f"debug_data/{endpoint}.json")
            path.parent.mkdir(parents=True, exist_ok=True)

            with path.open("w", encoding="utf-8") as f:
                json.dump(debug_data, f, indent=2)

        if req.status_code != 200:
            raise ApiError(**data)

        return model.model_validate(data)

    def getAlbum(self, album_id: str | int):
        return self.fetch(
            Album, f"albums/{album_id}", {"countryCode": self.country_code}
        )

    def getAlbumItems(self, album_id: str | int, limit=LIMITS.ALBUM_ITEMS, offset=0):
        return self.fetch(
            AlbumItems,
            f"albums/{album_id}/items",
            {
                "countryCode": self.country_code,
                "limit": ensureLimit(limit, self.LIMITS.ALBUM_ITEMS_MAX),
                "offset": offset,
            },
        )

    def getAlbumItemsCredits(
        self, album_id: str | int, limit=LIMITS.ALBUM_ITEMS, offset=0
    ):
        return self.fetch(
            AlbumItemsCredits,
            f"albums/{album_id}/items/credits",
            {
                "countryCode": self.country_code,
                "limit": ensureLimit(limit, self.LIMITS.ALBUM_ITEMS_MAX),
                "offset": offset,
            },
        )

    def getArtist(self, artist_id: str | int):
        return self.fetch(
            Artist,
            f"artists/{artist_id}",
            {"countryCode": self.country_code},
            expire_after=3600,
        )

    def getArtistAlbums(
        self,
        artist_id: str | int,
        limit=LIMITS.ARTIST_ALBUMS,
        offset=0,
        filter: Literal["ALBUMS", "EPSANDSINGLES"] = "ALBUMS",
    ):
        return self.fetch(
            ArtistAlbumsItems,
            f"artists/{artist_id}/albums",
            {
                "countryCode": self.country_code,
                "limit": limit,  # tested limit 10,000
                "offset": offset,
                "filter": filter,
            },
            expire_after=3600,
        )

    def getMix(
        self,
        mix_id: str | int,
        limit=LIMITS.MIX_ITEMS,
        offset=0,
    ):
        return self.fetch(
            MixItems,
            f"mixes/{mix_id}/items",
            {
                "countryCode": self.country_code,
                "limit": limit,
                "offset": offset,
            },
            expire_after=3600,
        )

    def getFavorites(self):
        return self.fetch(
            Favorites,
            f"users/{self.user_id}/favorites/ids",
            {"countryCode": self.country_code},
            expire_after=EXPIRE_IMMEDIATELY,
        )

    def getPlaylist(self, playlist_uuid: str):
        return self.fetch(
            Playlist,
            f"playlists/{playlist_uuid}",
            {"countryCode": self.country_code},
        )

    def getPlaylistItems(self, playlist_uuid: str, limit=LIMITS.PLAYLIST, offset=0):
        return self.fetch(
            PlaylistItems,
            f"playlists/{playlist_uuid}/items",
            {
                "countryCode": self.country_code,
                "limit": limit,
                "offset": offset,
            },
            expire_after=EXPIRE_IMMEDIATELY,
        )

    def getSearch(self, query: str):
        return self.fetch(
            Search,
            "search",
            {"countryCode": self.country_code, "query": query},
            expire_after=EXPIRE_IMMEDIATELY,
        )

    def getSession(self):
        return self.fetch(SessionResponse, "sessions", expire_after=DO_NOT_CACHE)

    def getLyrics(self, track_id: str | int):
        return self.fetch(
            Lyrics, f"tracks/{track_id}/lyrics", {"countryCode": self.country_code}
        )

    def getTrack(self, track_id: str | int):
        return self.fetch(
            Track, f"tracks/{track_id}", {"countryCode": self.country_code}
        )

    def getTrackStream(self, track_id: str | int, quality: TrackQuality):
        return self.fetch(
            TrackStream,
            f"tracks/{track_id}/playbackinfo",
            {
                "audioquality": quality,
                "playbackmode": "STREAM",
                "assetpresentation": "FULL",
            },
            expire_after=DO_NOT_CACHE,
        )

    def getVideo(self, video_id: str | int):
        return self.fetch(
            Video, f"videos/{video_id}", {"countryCode": self.country_code}
        )

    def getVideoStream(self, video_id: str | int):
        return self.fetch(
            VideoStream,
            f"videos/{video_id}/playbackinfo",
            {
                "videoquality": "HIGH",
                "playbackmode": "STREAM",
                "assetpresentation": "FULL",
            },
            expire_after=DO_NOT_CACHE,
        )



import logging
import click
import asyncio

from time import perf_counter
from concurrent.futures import ThreadPoolExecutor, Future
from pathlib import Path
from requests import Session

from rich.highlighter import ReprHighlighter
from rich.progress import (
    SpinnerColumn,
    Progress,
    TextColumn,
)

from tiddl.download import parseTrackStream, parseVideoStream
from tiddl.exceptions import ApiError, AuthError
from tiddl.metadata import Cover, addMetadata, addVideoMetadata
from tiddl.models.api import AlbumItemsCredits
from tiddl.models.constants import ARG_TO_QUALITY, TrackArg, SinglesFilter
from tiddl.models.resource import Track, Video, Album
from tiddl.utils import (
    TidalResource,
    formatResource,
    convertFileExtension,
    savePlaylistM3U,
    findTrackFilename,
)

from tiddl.cli.ctx import Context, passContext
from tiddl.cli.download.fav import FavGroup
from tiddl.cli.download.file import FileGroup
from tiddl.cli.download.search import SearchGroup
from tiddl.cli.download.url import UrlGroup

from typing import List, Union

logger = logging.getLogger(__name__)


@click.command("download")
@click.option(
    "--quality",
    "-q",
    "QUALITY",
    type=click.Choice(TrackArg.__args__),
    help="Track quality.",
)
@click.option(
    "--output",
    "-o",
    "TEMPLATE",
    type=str,
    help="Format output file template. "
    "This will be used instead of your config templates.",
)
@click.option(
    "--path",
    "-p",
    "PATH",
    type=str,
    help="Base path of download directory. Default is ~/Music/Tiddl.",
)
@click.option(
    "--threads",
    "-t",
    "THREADS_COUNT",
    type=int,
    help="Number of threads to use in concurrent download; use with caution.",
)
@click.option(
    "--noskip",
    "-ns",
    "DO_NOT_SKIP",
    is_flag=True,
    default=False,
    help="Do not skip already downloaded files.",
)
@click.option(
    "--singles",
    "-s",
    "SINGLES_FILTER",
    type=click.Choice(SinglesFilter.__args__),
    help="Defines how to treat artist EPs and singles, used while downloading artist.",
)
@click.option(
    "--lyrics",
    "-l",
    "EMBED_LYRICS",
    is_flag=True,
    help="Embed track lyrics in file metadata.",
)
@click.option(
    "--video",
    "-V",
    "DOWNLOAD_VIDEO",
    is_flag=True,
    help="Enable downloading videos",
)
@click.option(
    "--scan-path",
    "SCAN_PATH",
    type=str,
    help="Base directory to scan for existing tracks. Default is 'path'",
)
@click.option(
    "--save-m3u",
    "-m3u",
    "SAVE_M3U",
    is_flag=True,
    help="Save M3U file for playlists.",
)
@passContext
def DownloadCommand(
    ctx: Context,
    QUALITY: TrackArg | None,
    TEMPLATE: str | None,
    PATH: str | None,
    THREADS_COUNT: int | None,
    DO_NOT_SKIP: bool,
    SINGLES_FILTER: SinglesFilter,
    EMBED_LYRICS: bool,
    DOWNLOAD_VIDEO: bool,
    SCAN_PATH: str | None,
    SAVE_M3U: bool,
):
    """Download resources"""
    DOWNLOAD_VIDEO = DOWNLOAD_VIDEO or ctx.obj.config.download.download_video
    SINGLES_FILTER = SINGLES_FILTER or ctx.obj.config.download.singles_filter
    EMBED_LYRICS = EMBED_LYRICS or ctx.obj.config.download.embed_lyrics

    # TODO: pretty print
    logger.debug(
        (
            QUALITY,
            TEMPLATE,
            PATH,
            THREADS_COUNT,
            DO_NOT_SKIP,
            SINGLES_FILTER,
            EMBED_LYRICS,
            DOWNLOAD_VIDEO,
            SCAN_PATH,
            SAVE_M3U,
        )
    )

    DOWNLOAD_QUALITY = ARG_TO_QUALITY[QUALITY or ctx.obj.config.download.quality]

    api = ctx.obj.getApi()

    progress = Progress(
        SpinnerColumn(),
        TextColumn(
            "{task.description} • "
            "{task.fields[speed]:.2f} Mbps • {task.fields[size]:.2f} MB",
            highlighter=ReprHighlighter(),
        ),
        console=ctx.obj.console,
        transient=True,
        auto_refresh=True,
    )

    def handleItemDownload(
        item: Union[Track, Video],
        path: Path,
        cover_data=b"",
        credits: List[AlbumItemsCredits.ItemWithCredits.CreditsEntry] = [],
        album_artist="",
    ) -> Path:
        if isinstance(item, Track):
            track_stream = api.getTrackStream(item.id, quality=DOWNLOAD_QUALITY)
            description = (
                f"Track '{item.title}' "
                f"{(str(track_stream.bitDepth) + ' bit') if track_stream.bitDepth else ''} "
                f"{str(track_stream.sampleRate) + ' kHz' if track_stream.sampleRate else ''}"
            )

            urls, extension = parseTrackStream(track_stream)
        elif isinstance(item, Video):
            video_stream = api.getVideoStream(item.id)
            description = f"Video '{item.title}' {video_stream.videoQuality} quality"

            urls = parseVideoStream(video_stream)
            extension = ".ts"
        else:
            raise TypeError(
                f"Invalid item type: expected an instance of Track or Video, "
                f"received an instance of {type(item).__name__}. "
            )

        task_id = progress.add_task(
            description=description,
            start=True,
            visible=True,
            total=None,
            # fields
            speed=0,
            size=0,
        )

        with Session() as s:
            stream_data = b""
            time_start = perf_counter()

            for url in urls:
                req = s.get(url)

                assert req.status_code == 200, (
                    f"Could not download stream data for: "
                    f"{type(item).__name__} '{item.title}', "
                    f"status code: {req.status_code}"
                )

                stream_data += req.content
                speed = len(stream_data) / (perf_counter() - time_start) / (1024 * 128)
                size = len(stream_data) / 1024**2
                progress.update(
                    task_id,
                    advance=len(req.content),
                    speed=speed,
                    size=size,
                )

        path = path.with_suffix(extension)
        path.parent.mkdir(parents=True, exist_ok=True)

        with path.open("wb") as f:
            f.write(stream_data)

        if isinstance(item, Track):
            if not cover_data and item.album.cover:
                cover_data = Cover(
                    item.album.cover, size=ctx.obj.config.cover.size
                ).content

            lyrics_subtitles = ""

            if EMBED_LYRICS:
                try:
                    lyrics_subtitles = api.getLyrics(item.id).subtitles
                except Exception as e:
                    logger.error(e)

            if track_stream.audioQuality in ["HI_RES_LOSSLESS"]:
                path = asyncio.run(
                    convertFileExtension(
                        source_file=path,
                        extension=".flac",
                        remove_source=True,
                        is_video=False,
                        copy_audio=True,  # extract flac from m4a container
                    )
                )

            try:
                addMetadata(
                    path,
                    item,
                    cover_data,
                    credits,
                    album_artist=album_artist,
                    lyrics=lyrics_subtitles,
                )
            except Exception as e:
                logger.error(f"Can not add metadata to: {path}, {e}")

        elif isinstance(item, Video):
            path = asyncio.run(
                convertFileExtension(
                    source_file=path,
                    extension=".mp4",
                    remove_source=True,
                    is_video=True,
                    copy_audio=True,
                )
            )

            try:
                addVideoMetadata(path, item)
            except Exception as e:
                logger.error(f"Can not add metadata to: {path}, {e}")

        progress.remove_task(task_id)
        logger.info(f"{item.title!r} • {speed:.2f} Mbps • {size:.2f} MB")

        return path

    pool = ThreadPoolExecutor(
        max_workers=THREADS_COUNT or ctx.obj.config.download.threads
    )

    def submitItem(
        item: Union[Track, Video],
        filename: str,
        cover_data=b"",
        credits: List[AlbumItemsCredits.ItemWithCredits.CreditsEntry] = [],
        album_artist="",
    ) -> Future[Path] | None:
        if not item.allowStreaming:
            logger.warning(
                f"✖ {type(item).__name__} '{item.title}' does not allow streaming"
            )
            return

        download_path = Path(PATH) if PATH else ctx.obj.config.download.path
        download_path /= f"{filename}.*"

        scan_path = Path(SCAN_PATH) if SCAN_PATH else ctx.obj.config.download.scan_path
        if scan_path:
            scan_path /= f"{filename}.*"
        else:
            scan_path = download_path

        # Respect DOWNLOAD_VIDEO = FALSE over DO_NOT_SKIP (as it's for the file exists check)
        if isinstance(item, Video) and not DOWNLOAD_VIDEO:
            logger.warning(f"Video '{item.title}' skipped as DOWNLOAD_VIDEO is false")
            return

        # check if item is already downloaded (unless DO_NOT_SKIP is set, then override anything)
        if not DO_NOT_SKIP:
            if isinstance(item, Track):
                track_path = findTrackFilename(
                    item.audioQuality, DOWNLOAD_QUALITY, scan_path
                )
                if track_path.exists():
                    logger.info(f"Track '{item.title}' skipped - exists")
                    future = Future()
                    future.set_result(track_path)
                    return future

            elif isinstance(item, Video):
                if scan_path.with_suffix(".mp4").exists():
                    logger.info(f"Video '{item.title}' skipped - exists")
                    return

        future = pool.submit(
            handleItemDownload,
            item=item,
            path=download_path,
            cover_data=cover_data,
            credits=credits,
            album_artist=album_artist,
        )

        return future

    def downloadAlbum(album: Album):
        logger.info(f"Album {album.title!r}")

        cover = (
            Cover(uid=album.cover, size=ctx.obj.config.cover.size)
            if album.cover
            else None
        )
        is_cover_saved = False

        offset = 0

        while True:
            album_items = api.getAlbumItemsCredits(album.id, offset=offset)

            for item in album_items.items:
                filename = formatResource(
                    template=TEMPLATE or ctx.obj.config.template.album,
                    resource=item.item,
                    album_artist=album.artist.name,
                )

                if cover and not is_cover_saved and ctx.obj.config.cover.save:
                    path = Path(PATH) if PATH else ctx.obj.config.download.path
                    cover_path = path / Path(filename).parent
                    cover.save(cover_path, ctx.obj.config.cover.filename)
                    is_cover_saved = True

                submitItem(
                    item.item,
                    filename,
                    cover.content if cover else b"",
                    item.credits,
                    album.artist.name,
                )

            if album_items.limit + album_items.offset > album_items.totalNumberOfItems:
                break

            offset += album_items.limit

    def handleResource(resource: TidalResource) -> None:
        logger.debug(f"'{resource}'")

        match resource.type:
            case "track":
                track = api.getTrack(resource.id)
                filename = formatResource(
                    TEMPLATE or ctx.obj.config.template.track, track
                )

                submitItem(track, filename)

            case "video":
                video = api.getVideo(resource.id)
                filename = formatResource(
                    TEMPLATE or ctx.obj.config.template.video, video
                )

                submitItem(video, filename)

            case "album":
                album = api.getAlbum(resource.id)

                downloadAlbum(album)

            case "mix":
                mix = api.getMix(resource.id)

                for mix_item in mix.items:
                    filename = formatResource(
                        TEMPLATE or ctx.obj.config.template.track, mix_item.item
                    )

                    submitItem(mix_item.item, filename)

            case "artist":
                artist = api.getArtist(resource.id)
                logger.info(f"Artist {artist.name!r}")

                def getAllAlbums(singles: bool):
                    offset = 0

                    while True:
                        artist_albums = api.getArtistAlbums(
                            resource.id,
                            offset=offset,
                            filter="EPSANDSINGLES" if singles else "ALBUMS",
                        )

                        for album in artist_albums.items:
                            downloadAlbum(album)

                        if (
                            artist_albums.limit + artist_albums.offset
                            > artist_albums.totalNumberOfItems
                        ):
                            break

                        offset += artist_albums.limit

                if SINGLES_FILTER == "include":
                    getAllAlbums(False)
                    getAllAlbums(True)
                else:
                    getAllAlbums(SINGLES_FILTER == "only")

            case "playlist":
                playlist = api.getPlaylist(resource.id)
                logger.info(f"downloading playlist {playlist.title!r}")
                offset = 0
                playlist_path = None

                futures: list[tuple[Future[Path], Track]] = []

                while True:
                    playlist_items = api.getPlaylistItems(playlist.uuid, offset=offset)

                    for item in playlist_items.items:
                        filename = formatResource(
                            template=TEMPLATE or ctx.obj.config.template.playlist,
                            resource=item.item,
                            playlist_title=playlist.title,
                            playlist_index=item.item.index // 100000,
                        )

                        future = submitItem(item.item, filename)
                        if future:
                            futures.append((future, item.item))

                        playlist_path = Path(filename).parent

                    if (
                        playlist_items.limit + playlist_items.offset
                        > playlist_items.totalNumberOfItems
                    ):
                        break

                    offset += playlist_items.limit

                playlist_tracks: list[tuple[Path, Track]] = []
                for future, track in futures:
                    track_path = future.result()
                    playlist_tracks.append((track_path, track))

                path = Path(PATH) if PATH else ctx.obj.config.download.path

                if playlist_path and (
                    SAVE_M3U or ctx.obj.config.download.save_playlist_m3u
                ):
                    savePlaylistM3U(
                        playlist_tracks=playlist_tracks,
                        path=path / playlist_path,
                        filename=f"{playlist.title}.m3u",
                    )

                if playlist.squareImage and playlist_path:
                    cover = Cover(
                        uid=playlist.squareImage,
                        size=1080,  # playlist cover must be 1080x1080
                    )
                    cover.save(path / playlist_path, ctx.obj.config.cover.filename)

    progress.start()

    # TODO: make sure every resource is unique
    for resource in ctx.obj.resources:
        try:
            handleResource(resource)

        except AuthError as e:
            logger.error(e)
            break

        except ApiError as e:
            logger.error(e)

            # session does not have streaming privileges
            if e.sub_status == 4006:
                break

    pool.shutdown(wait=True)
    progress.stop()


UrlGroup.add_command(DownloadCommand)
SearchGroup.add_command(DownloadCommand)
FavGroup.add_command(DownloadCommand)
FileGroup.add_command(DownloadCommand)



from os import environ, makedirs
from pydantic import BaseModel
from pathlib import Path

from tiddl.models.constants import TrackArg, SinglesFilter

TIDDL_ENV_KEY = "TIDDL_PATH"

# 3.0 TODO: rename HOME_PATH to TIDDL_PATH
# 3.0 TODO: add /tiddl to Path.home()
HOME_PATH = Path(environ[TIDDL_ENV_KEY]) if environ.get(TIDDL_ENV_KEY) else Path.home()

makedirs(HOME_PATH, exist_ok=True)

CONFIG_PATH = HOME_PATH / "tiddl.json"
CONFIG_INDENT = 2


class TemplateConfig(BaseModel):
    track: str = "{artist} - {title}"
    video: str = "{artist} - {title}"
    album: str = "{album_artist}/{album}/{number:02d}. {title}"
    playlist: str = "{playlist}/{playlist_number:02d}. {artist} - {title}"


class DownloadConfig(BaseModel):
    quality: TrackArg = "high"
    path: Path = Path.home() / "Music" / "Tiddl"
    threads: int = 4
    singles_filter: SinglesFilter = "none"
    embed_lyrics: bool = False
    download_video: bool = False
    scan_path: Path | None = path
    save_playlist_m3u: bool = False


class AuthConfig(BaseModel):
    token: str = ""
    refresh_token: str = ""
    expires: int = 0
    user_id: str = ""
    country_code: str = ""


class CoverConfig(BaseModel):
    save: bool = False
    size: int = 1280
    filename: str = "cover.jpg"


class Config(BaseModel):
    template: TemplateConfig = TemplateConfig()
    download: DownloadConfig = DownloadConfig()
    cover: CoverConfig = CoverConfig()
    auth: AuthConfig = AuthConfig()
    omit_cache: bool = False

    def save(self):
        with open(CONFIG_PATH, "w") as f:
            f.write(self.model_dump_json(indent=CONFIG_INDENT))

    @classmethod
    def fromFile(cls):
        try:
            with CONFIG_PATH.open() as f:
                config = cls.model_validate_json(f.read())
        except FileNotFoundError:
            config = cls()

        config.save()
        return config



import click
import logging

from rich.logging import RichHandler

from tiddl.config import HOME_PATH
from tiddl.cli.ctx import ContextObj, passContext, Context
from tiddl.cli.auth import AuthGroup
from tiddl.cli.download import UrlGroup, FavGroup, SearchGroup, FileGroup
from tiddl.cli.config import ConfigCommand
from tiddl.cli.auth import refresh


@click.group()
@passContext
@click.option("--verbose", "-v", is_flag=True, help="Show debug logs.")
@click.option("--quiet", "-q", is_flag=True, help="Suppress logs.")
@click.option(
    "--no-cache", "-nc", is_flag=True, help="Omit Tidal API requests caching."
)
def cli(ctx: Context, verbose: bool, quiet: bool, no_cache: bool):
    """TIDDL - Tidal Downloader \u266b"""
    ctx.obj = ContextObj()

    # latest logs
    file_handler = logging.FileHandler(
        HOME_PATH / "tiddl.log", mode="w", encoding="utf-8"
    )

    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(
        logging.Formatter(
            "%(levelname)s [%(name)s.%(funcName)s] %(message)s", datefmt="[%X]"
        )
    )

    LEVEL = logging.DEBUG if verbose else logging.ERROR if quiet else logging.INFO

    rich_handler = RichHandler(console=ctx.obj.console, rich_tracebacks=True)
    rich_handler.setLevel(LEVEL)

    if LEVEL == logging.DEBUG:
        rich_handler.setFormatter(
            logging.Formatter("[%(name)s.%(funcName)s] %(message)s", datefmt="[%X]")
        )

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[
            rich_handler,
            file_handler,
        ],
        format="%(message)s",
        datefmt="[%X]",
    )

    logging.getLogger("urllib3").setLevel(logging.ERROR)

    if ctx.invoked_subcommand in ("fav", "file", "search", "url"):
        ctx.invoke(refresh)

    ctx.obj.initApi(omit_cache=no_cache)


cli.add_command(ConfigCommand)
cli.add_command(AuthGroup)
cli.add_command(UrlGroup)
cli.add_command(FavGroup)
cli.add_command(SearchGroup)
cli.add_command(FileGroup)

if __name__ == "__main__":
    cli()



import functools
import click

from rich.console import Console

from typing import Callable, TypeVar, cast

from tiddl.api import TidalApi
from tiddl.config import Config
from tiddl.utils import TidalResource


class ContextObj:
    api: TidalApi | None
    config: Config
    resources: list[TidalResource]
    console: Console

    def __init__(self) -> None:
        self.config = Config.fromFile()
        self.resources = []
        self.api = None
        self.console = Console()

    def initApi(self, omit_cache=False):
        auth = self.config.auth

        if auth.token and auth.user_id and auth.country_code:
            self.api = TidalApi(
                auth.token,
                auth.user_id,
                auth.country_code,
                omit_cache=omit_cache or self.config.omit_cache,
            )

    def getApi(self) -> TidalApi:
        if self.api is None:
            raise click.UsageError("You must login first")

        return self.api


class Context(click.Context):
    obj: ContextObj


F = TypeVar("F", bound=Callable[..., None])


def passContext(func: F) -> F:
    """Wrapper for @click.pass_context to use custom Context"""

    @click.pass_context
    @functools.wraps(func)
    def wrapper(ctx: click.Context, *args, **kwargs):
        custom_ctx = cast(Context, ctx)
        return func(custom_ctx, *args, **kwargs)

    return cast(F, wrapper)



import re
import os
import logging

from ffmpeg_asyncio import FFmpeg
from ffmpeg_asyncio.types import Option as FFmpegOption

from pydantic import BaseModel
from urllib.parse import urlparse
from pathlib import Path

from typing import Literal, Union, get_args

from tiddl.models.constants import TrackQuality, QUALITY_TO_ARG
from tiddl.models.resource import Track, Video

ResourceTypeLiteral = Literal["track", "video", "album", "playlist", "artist", "mix"]


class TidalResource(BaseModel):
    type: ResourceTypeLiteral
    id: str

    @property
    def url(self) -> str:
        return f"https://listen.tidal.com/{self.type}/{self.id}"

    @classmethod
    def fromString(cls, string: str):
        """
        Extracts the resource type (e.g., "track", "album")
        and resource ID from a given input string.

        The input string can either be a full URL or a shorthand string
        in the format `resource_type/resource_id` (e.g., `track/12345678`).
        """

        path = urlparse(string).path
        resource_type, resource_id = path.split("/")[-2:]

        if resource_type not in get_args(ResourceTypeLiteral):
            raise ValueError(f"Invalid resource type: {resource_type}")

        digit_resource_types: list[ResourceTypeLiteral] = [
            "track",
            "album",
            "video",
            "artist",
        ]

        if resource_type in digit_resource_types and not resource_id.isdigit():
            raise ValueError(f"Invalid resource id: {resource_id}")

        return cls(type=resource_type, id=resource_id)  # type: ignore

    def __str__(self) -> str:
        return f"{self.type}/{self.id}"


def sanitizeString(string: str) -> str:
    pattern = r'[\\/:"*?<>|]+'
    return re.sub(pattern, "", string)


def formatTrack(
    template: str,
    track: Track,
    album_artist="",
    playlist_title="",
    playlist_index=0,
) -> str:
    artist = sanitizeString(track.artist.name) if track.artist else ""
    features = [
        sanitizeString(track_artist.name)
        for track_artist in track.artists
        if track_artist.name != artist
    ]

    track_dict = {
        "id": str(track.id),
        "title": sanitizeString(track.title),
        "version": sanitizeString(track.version or ""),
        "artist": artist,
        "artists": ", ".join(features + [artist]),
        "features": ", ".join(features),
        "album": sanitizeString(track.album.title),
        "number": track.trackNumber,
        "disc": track.volumeNumber,
        "date": (track.streamStartDate if track.streamStartDate else ""),
        # i think we can remove year as we are able to format date
        "year": track.streamStartDate.strftime("%Y") if track.streamStartDate else "",
        "playlist": sanitizeString(playlist_title),
        "bpm": track.bpm or "",
        "quality": QUALITY_TO_ARG[track.audioQuality],
        "album_artist": sanitizeString(album_artist),
        "playlist_number": playlist_index or 0,
    }

    formatted_track = template.format(**track_dict).strip()

    disallowed_chars = r'[\\:"*?<>|]+'
    invalid_chars = re.findall(disallowed_chars, formatted_track)

    if invalid_chars:
        raise ValueError(
            f"Template '{template}' and formatted track '{formatted_track}' contains disallowed characters: {' '.join(sorted(set(invalid_chars)))}"
        )

    return formatted_track


def formatResource(
    template: str,
    resource: Union[Track, Video],
    album_artist="",
    playlist_title="",
    playlist_index=0,
) -> str:
    artist = sanitizeString(resource.artist.name) if resource.artist else ""

    features = [
        sanitizeString(item_artist.name)
        for item_artist in resource.artists
        if item_artist.name != artist
    ]

    resource_dict = {
        "id": str(resource.id),
        "title": sanitizeString(resource.title),
        "artist": artist,
        "artists": ", ".join(features + [artist]),
        "features": ", ".join(features),
        "album": sanitizeString(resource.album.title if resource.album else ""),
        "album_id": str(resource.album.id if resource.album else ""),
        "number": resource.trackNumber,
        "disc": resource.volumeNumber,
        "date": (resource.streamStartDate if resource.streamStartDate else ""),
        # i think we can remove year as we are able to format date
        "year": (
            resource.streamStartDate.strftime("%Y") if resource.streamStartDate else ""
        ),
        "playlist": sanitizeString(playlist_title),
        "album_artist": sanitizeString(album_artist),
        "playlist_number": playlist_index or 0,
        "quality": "",
        "version": "",
        "bpm": "",
    }

    if isinstance(resource, Track):
        resource_dict.update(
            {
                "version": sanitizeString(resource.version or ""),
                "quality": QUALITY_TO_ARG[resource.audioQuality],
                "bpm": resource.bpm or "",
            }
        )

    elif isinstance(resource, Video):
        resource_dict.update({"quality": resource.quality})

    formatted_template = template.format(**resource_dict).strip()

    disallowed_chars = r'[\\:"*?<>|]+'
    invalid_chars = re.findall(disallowed_chars, formatted_template)

    if invalid_chars:
        raise ValueError(
            f"Template '{template}' and formatted resource '{formatted_template}'"
            f"contains disallowed characters: {' '.join(sorted(set(invalid_chars)))}"
        )

    return formatted_template


def findTrackFilename(
    track_quality: TrackQuality, download_quality: TrackQuality, file_name: Path
) -> Path:
    """
    Predict track extension.
    """

    FLAC_QUALITIES: list[TrackQuality] = ["LOSSLESS", "HI_RES_LOSSLESS"]

    if download_quality in FLAC_QUALITIES and track_quality in FLAC_QUALITIES:
        extension = ".flac"
    else:
        extension = ".m4a"

    full_file_name = file_name.with_suffix(extension)

    return full_file_name


async def convertFileExtension(
    source_file: Path,
    extension: str,
    remove_source=False,
    is_video=False,
    copy_audio=False,
) -> Path:
    """
    Converts `source_file` extension and returns `Path` of file with new `extension`.

    Removes `source_file` when `remove_source` is truthy.
    """

    try:
        output_file = source_file.with_suffix(extension)
    except ValueError as e:
        logging.error(e)
        return source_file

    logging.debug((source_file, output_file, extension, copy_audio, is_video))

    if extension == source_file.suffix:
        logging.debug("Conversion not required, already %s", extension)
        return source_file

    ffmpeg_args: dict[str, FFmpegOption | None] = {"loglevel": "error"}

    if copy_audio:
        ffmpeg_args["acodec"] = "copy"

    if is_video:
        ffmpeg_args["vcodec"] = "copy"

    try:
        logging.debug("Trying conversion")
        ffmpeg = FFmpeg().option("y")
        ffmpeg.input(str(source_file))
        ffmpeg.output(str(output_file), ffmpeg_args)

        @ffmpeg.on("completed")
        def on_completed():
            logging.debug(f"converted {output_file}")
            if remove_source:
                try:
                    os.remove(source_file)
                except OSError as e:
                    logging.error(f"can't remove source file {source_file}: {e}")

        await ffmpeg.execute()

    except Exception as e:
        logging.error(f"can't convert file {source_file}: {e}")
        return source_file

    return output_file


def savePlaylistM3U(
    playlist_tracks: list[tuple[Path, Track]], path: Path, filename="playlist.m3u"
):
    """
    playlist_tracks: [track_path, Track]
    path: m3u file location
    filename: name of the m3u file
    """

    file = path / sanitizeString(filename)
    logging.debug(f"saving m3u file at {file}")

    if not playlist_tracks:
        logging.warning(f"playlist {file} is empty")
        return

    try:
        with file.open("w", encoding="utf-8") as f:
            f.write("#EXTM3U\n")
            for track_path, track in playlist_tracks:
                f.write(
                    f"#EXTINF:{track.duration},{track.artist.name if track.artist else ''} - {track.title}\n{track_path}\n"
                )

            logging.debug(
                f"saved m3u file as {file} with {len(playlist_tracks)} tracks"
            )

    except Exception as e:
        logging.error(f"can't save playlist m3u file: {e}")



import logging
import requests

from os import makedirs
from pathlib import Path

from mutagen.easymp4 import EasyMP4 as MutagenEasyMP4
from mutagen.flac import FLAC as MutagenFLAC
from mutagen.flac import Picture
from mutagen.mp4 import MP4 as MutagenMP4
from mutagen.mp4 import MP4Cover

from tiddl.models.resource import Track, Video
from tiddl.models.api import AlbumItemsCredits

from typing import List

logger = logging.getLogger(__name__)


def addMetadata(
    track_path: Path,
    track: Track,
    cover_data=b"",
    credits: List[AlbumItemsCredits.ItemWithCredits.CreditsEntry] = [],
    album_artist="",
    lyrics="",
):
    logger.debug((track_path, track.id))

    extension = track_path.suffix

    # TODO: handle mutagen exceptions

    if extension == ".flac":
        metadata = MutagenFLAC(track_path)
        if cover_data:
            picture = Picture()
            picture.data = cover_data
            picture.mime = "image/jpeg"
            metadata.add_picture(picture)

        metadata["TITLE"] = track.title + (
            " ({})".format(track.version) if track.version else ""
        )
        metadata["WORK"] = track.title + (
            " ({})".format(track.version) if track.version else ""
        )
        metadata["TRACKNUMBER"] = str(track.trackNumber)
        metadata["DISCNUMBER"] = str(track.volumeNumber)

        metadata["ALBUM"] = track.album.title

        metadata["ARTIST"] = "; ".join(
            [artist.name.strip() for artist in track.artists]
        )

        if album_artist:
            metadata["ALBUMARTIST"] = album_artist
        elif track.artist:
            metadata["ALBUMARTIST"] = track.artist.name

        if track.streamStartDate:
            metadata["DATE"] = track.streamStartDate.strftime("%Y-%m-%d")
            metadata["ORIGINALDATE"] = track.streamStartDate.strftime("%Y-%m-%d")
            metadata["YEAR"] = str(track.streamStartDate.strftime("%Y"))
            metadata["ORIGINALYEAR"] = str(track.streamStartDate.strftime("%Y"))

        if track.copyright:
            metadata["COPYRIGHT"] = track.copyright

        metadata["ISRC"] = track.isrc

        if track.bpm:
            metadata["BPM"] = str(track.bpm)

        for entry in credits:
            metadata[entry.type.upper()] = [
                contributor.name for contributor in entry.contributors
            ]

        if lyrics:
            metadata["LYRICS"] = lyrics

    elif extension == ".m4a":
        if lyrics or cover_data:
            metadata = MutagenMP4(track_path)

            if lyrics:
                metadata["\xa9lyr"] = [lyrics]

            if cover_data:
                metadata["covr"] = [
                    MP4Cover(cover_data, imageformat=MP4Cover.FORMAT_JPEG)
                ]

            metadata.save()

        metadata = MutagenEasyMP4(track_path)
        metadata.update(
            {
                "title": track.title,
                "tracknumber": str(track.trackNumber),
                "discnumber": str(track.volumeNumber),
                "copyright": track.copyright if track.copyright else "",
                "albumartist": track.artist.name if track.artist else "",
                "artist": ";".join([artist.name.strip() for artist in track.artists]),
                "album": track.album.title,
                "date": str(track.streamStartDate) if track.streamStartDate else "",
                "bpm": str(track.bpm or 0),
            }
        )

    else:
        raise ValueError(f"Unknown file extension: {extension}")

    try:
        metadata.save(track_path)
    except Exception as e:
        logger.error(f"Failed to add metadata to {track_path}: {e}")


def addVideoMetadata(path: Path, video: Video):
    metadata = MutagenEasyMP4(path)

    metadata.update(
        {
            "title": video.title,
            "albumartist": video.artist.name if video.artist else "",
            "artist": ";".join([artist.name.strip() for artist in video.artists]),
            "album": video.album.title if video.album else "",
            "date": str(video.streamStartDate) if video.streamStartDate else "",
        }
    )

    if video.trackNumber:
        metadata["tracknumber"] = str(video.trackNumber)

    if video.volumeNumber:
        metadata["discnumber"] = str(video.volumeNumber)

    try:
        metadata.save(path)
    except Exception as e:
        logger.error(f"Failed to add metadata to {path}: {e}")


class Cover:
    # TODO: cache covers

    def __init__(self, uid: str, size=1280) -> None:
        if size > 1280:
            logger.warning(
                f"can not set cover size higher than 1280 (user set: {size})"
            )
            size = 1280

        self.uid = uid

        formatted_uid = uid.replace("-", "/")
        self.url = (
            f"https://resources.tidal.com/images/{formatted_uid}/{size}x{size}.jpg"
        )

        logger.debug((self.uid, self.url))

        self.content = self._get()

    def _get(self) -> bytes:
        req = requests.get(self.url)

        if req.status_code != 200:
            logger.error(f"could not download cover. ({req.status_code}) {self.url}")
            return b""

        logger.debug(f"got cover: {self.uid}")

        return req.content

    def save(self, directory_path: Path, filename="cover.jpg"):
        if not self.content:
            logger.error("cover file content is empty")
            return

        file = directory_path / filename

        if file.exists():
            logger.debug(f"cover already exists ({file})")
            return

        makedirs(directory_path, exist_ok=True)

        try:
            with file.open("wb") as f:
                f.write(self.content)

        except FileNotFoundError as e:
            logger.error(f"could not save cover. {file} -> {e}")



import click

from tiddl.utils import TidalResource, ResourceTypeLiteral
from tiddl.cli.ctx import Context, passContext

ResourceTypeList: list[ResourceTypeLiteral] = [
    "track",
    "video",
    "album",
    "artist",
    "playlist",
]


@click.group("fav")
@click.option(
    "--resource",
    "-r",
    "resource_types",
    multiple=True,
    type=click.Choice(ResourceTypeList),
)
@passContext
def FavGroup(ctx: Context, resource_types: list[ResourceTypeLiteral]):
    """Get your Tidal favorites."""

    api = ctx.obj.getApi()

    favorites = api.getFavorites()
    favorites_dict = favorites.model_dump()

    click.echo(type(resource_types))

    if not resource_types:
        resource_types = ResourceTypeList

    stats: dict[ResourceTypeLiteral, int] = dict()

    for resource_type in resource_types:
        resources = favorites_dict[resource_type.upper()]

        stats[resource_type] = len(resources)

        for resource_id in resources:
            ctx.obj.resources.append(TidalResource(id=resource_id, type=resource_type))

    # TODO: show pretty message

    click.echo(click.style(f"Loaded {len(ctx.obj.resources)} resources", "green"))

    for resource_type, count in stats.items():
        click.echo(f"{resource_type} - {count}")



import click

from tiddl.utils import TidalResource
from tiddl.models.resource import Artist, Album, Playlist, Track, Video
from tiddl.cli.ctx import Context, passContext


@click.group("search")
@click.argument("query")
@passContext
def SearchGroup(ctx: Context, query: str):
    """Search on Tidal."""

    # TODO: give user interactive choice what to select

    api = ctx.obj.getApi()

    search = api.getSearch(query)

    # issue is that we get resource data in search api call,
    # in download we refetch that data.
    # it's not that big deal as we refetch one resource at most,
    # but it should be redesigned

    if not search.topHit:
        click.echo(f"No search results for '{query}'")
        return

    value = search.topHit.value
    icon = click.style("\u2bcc", "magenta")

    if isinstance(value, Album):
        resource = TidalResource(type="album", id=str(value.id))
        click.echo(f"{icon} Album {value.title}")
    elif isinstance(value, Artist):
        resource = TidalResource(type="artist", id=str(value.id))
        click.echo(f"{icon} Artist {value.name}")
    elif isinstance(value, Track):
        resource = TidalResource(type="track", id=str(value.id))
        click.echo(f"{icon} Track {value.title}")
    elif isinstance(value, Playlist):
        resource = TidalResource(type="playlist", id=str(value.uuid))
        click.echo(f"{icon} Playlist {value.title}")
    elif isinstance(value, Video):
        resource = TidalResource(type="video", id=str(value.id))
        click.echo(f"{icon} Video {value.title}")

    ctx.obj.resources.append(resource)



import click
import logging

from time import sleep, time

from tiddl.config import AuthConfig
from tiddl.auth import (
    getDeviceAuth,
    getToken,
    refreshToken,
    removeToken,
    AuthError,
)
from tiddl.cli.ctx import passContext, Context


logger = logging.getLogger(__name__)


@click.group("auth")
def AuthGroup():
    """Manage Tidal token."""


@AuthGroup.command("refresh")
@passContext
def refresh(ctx: Context):
    """Refresh auth token when is expired"""

    logger.debug("Invoked refresh command")

    auth = ctx.obj.config.auth

    if auth.refresh_token and time() > auth.expires:
        logger.info("Refreshing token...")
        token = refreshToken(auth.refresh_token)

        ctx.obj.config.auth.expires = token.expires_in + int(time())
        ctx.obj.config.auth.token = token.access_token

        ctx.obj.config.save()
        logger.info("Refreshed auth token!")


@AuthGroup.command("login")
@passContext
def login(ctx: Context):
    """Add token to the config"""

    logger.debug("Invoked login command")

    if ctx.obj.config.auth.token:
        logger.info("Already logged in.")
        ctx.invoke(refresh)
        return

    auth = getDeviceAuth()

    uri = f"https://{auth.verificationUriComplete}"
    click.launch(uri)

    logger.info(f"Go to {uri} and complete authentication!")

    auth_end_at = time() + auth.expiresIn

    while True:
        sleep(auth.interval)

        try:
            token = getToken(auth.deviceCode)
        except AuthError as e:
            if e.error == "authorization_pending":
                time_left = auth_end_at - time()
                minutes, seconds = time_left // 60, int(time_left % 60)

                click.echo(f"\rTime left: {minutes:.0f}:{seconds:02d}", nl=False)
                continue

            if e.error == "expired_token":
                logger.info("\nTime for authentication has expired.")
                break

        ctx.obj.config.auth = AuthConfig(
            token=token.access_token,
            refresh_token=token.refresh_token,
            expires=token.expires_in + int(time()),
            user_id=str(token.user.userId),
            country_code=token.user.countryCode,
        )
        ctx.obj.config.save()

        logger.info("\nAuthenticated!")

        break


@AuthGroup.command("logout")
@passContext
def logout(ctx: Context):
    """Remove token from config"""

    logger.debug("Invoked logout command")

    access_token = ctx.obj.config.auth.token

    if not access_token:
        logger.info("Not logged in.")
        return

    removeToken(access_token)

    ctx.obj.config.auth = AuthConfig()
    ctx.obj.config.save()

    logger.info("Logged out!")
